/*  mergefour.c -- Merge a two-by-two array of overlapping images

NAME:
    MERGEFOUR

PURPOSE:
    MERGEFOUR uses linear weighting functions to seamlessly merge a
    two-by-two array of overlapping images which typically represent the
    northwest, northeast, southwest, and southeast quandrants of a
    region of interest.

ARGUMENTS:
    IN		Names of the four input images.
    OUT		Output image.
    ODTYPE	Output data type.

FILE REFERENCES:
    IN		READ to get input pixels to be merged.
    OUT		CREATEd to receive merged output.

PROGRAM HISTORY:                                  
    2000 May	Initial version, R. A. White, PSU/ESSC.
    20o5 Mar	Updated for LAS 7.4, Linux.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.X and Linux 2.4.

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in mergefour.pdf.

EXTERNAL FUNCTIONS CALLED:

(Program Specific Functions)

    mrf_getarg	Get program arguments.
    jff_getwnd	Determine locations of inputs relative to output.
    mrf_defreg	Define overlap regions.
    mrf_merge	Merge inputs into output image.

ALGORITHM DESCRIPTION:
    Initialize TAE interface <c_lasini>
    Get program arguments <mrf_getarg>
    Determine locations of inputs relative to output <mrf_getwnd>
    Define overlap regions and compute weights <mrf_defrgn>
    Merge inputs into output image <mrf_merge>
    Build non-duplicated list of input image names
    Create/update history  file <c_hstout>
    Output completion message <c_complt>
    									
***************   End of Prolog for MERGEFOUR   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O functions		*/
#include	<string.h>	/* String manipulation functions	*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"util.h"	/* LAS utility function definitions	*/
#include	"mergefour.h"	/* Program-specifiec definitions	*/
									/*
***************   Main Routine for MERGEFOUR   ***************	*/

int main()
    {
	                                              			/*
Local Variable Definitions:						*/
    struct MRF_IMG
		input[4];	/* Input image specifications */
    long	ifminmax;	/* Flag for output minima/maxima */
    long	i;		/* Loop counter/index */
    char	innames[MAXIMG][CMLEN];/* List of all input images */
    long	inseq[4];	/* Sequence for reading inputs */
    long	inwind[4][4];	/* Window specification for each input
				   image relative to output image */
    long	j;		/* Loop counter/index */
    long	k;		/* Loop counter/index */
    long	odtype;		/* Output data type code */
    long	one = 1;	/* Constant value */
    struct DDR	outddr;		/* Georeferencing information for output
				   image */
    char	outimg[CMLEN];	/* Name of output image file */
    long	pmsgtim;	/* Interval between progress messages */
    struct MRF_REGION
		region[49];	/* Specifications for each overlap
				   region */
    long	stdlun;		/* Unit number of stdout (not ref'd) */
    long	totin;		/* Total number of input images */
    struct PARBLK
		vblock;		/* TAE parameter block */


/*  Get program arguments from TAE interface */

    c_inlas (&vblock, &stdlun);
    mrf_getarg (&vblock, input, outimg, &odtype, &pmsgtim, &ifminmax);

/*  Determine locations of inputs relative to output  */

    mrf_getwnd (input, inseq, &outddr, inwind);

/*  Define overlap regions and compute weights  */

    mrf_defrgn (&outddr, inwind, region);

/*  Merge inputs into output image  */

    if (mrf_merge (input, pmsgtim, inseq, inwind, region, outimg,
      &outddr, &odtype) != E_SUCC)
        c_clean (&vblock, "mergefour", (void *) outimg, &one);

/*  Build non-duplicated list of input image names  */

    for (i = 0, totin = 0; i < 4; i++)
	for (j = 0; j < input[i].totimage; j++)
	    {
	    for (k = 0; k < totin; k++)
		if (! strcmp (input[i].hname[j], innames[k]))
		    break;
	    if (k >= totin)
		strcpy( innames[totin++], input[i].hname[j] );
	    }

/*  Create/update history file  */

    c_hstout (&vblock, outimg, innames, &totin, "");

/*  Output completion message <c_complt>  */

    c_complt ("MERGEFOUR", (void *) outimg, &one);

    return 0;	/* To suppress compiler warning */

    }			/* End of Program MERGEFOUR */
