/*  mrf_defrgn.c -- Define boundaries and weights for overlap regions 

NAME:
    mrf_defrgn

PURPOSE:
    Called by MERGEFOUR to define the boundaries of all overlap regions
    between the four input images and initialize the weigths used for
    averaging input pixels within each region.

ARGUMENTS:

    struct DDR
	    *outddr	(IN)	DDR for output image.
    long    inwind[][4]	(IN)	Input image window locations relative to
				output image.
    struct MRF_REGION
	    *region	(OUT)	Specifications for each overlap region.

RETURN VALUE:  
    E_SUCC	Successful completion.
		(Function aborts on error)

PROGRAM HISTORY:                                  
    2000-05	Initial version, R. A. White, PSU/ESSC.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.X and Linux 2.4.

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in MERGEFOUR.PDF.

ALGORITHM DESCRIPTION:

    Sort region terminating lines and samples N to S and W to E
    For each region
	Set boundaries of region
	For each input image window
	    If input image window contains region
		Determine distances from region ulc to window edges
		Compute relative weights for x and y directions one line
		  and sample outside each corner of region
	    Else
	        Set relative weights to zero
	For each corner of region
	    If non-zero x-weights for  both images in upper row or in
	      lower row (corner not on left/right edge of either image)
		Set y weights equal
	    If non-zero y-weights for both images in left column or in
	      right column (corner not on top/botton edge of image)
		Set x weights equal
	    Compute normalized weights at region corner
	Compute coefficients of bilinear weight interpolation
    									
***************   End of Prolog for mrf_defrgn   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O definitions		*/
#include	<string.h>	/* String manipulation functions	*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"mergefour.h"	/* Program-specific definitions		*/
									/*
Local Function Prototypes:						*/

    int compare_edge ( const void *edge1, const void *edge2);
									/*
***************   Start of Function mrf_defrgn   ***************	*/

FUNCTION long mrf_defrgn(

    struct DDR
	    *outddr,	/* DDR for output image */
    long    inwind[][4], /* Input image window locations relative to
			   output image */
    struct MRF_REGION
	    *region)	/* Specifications for each overlap region */
    {
	                                              			/*
Local Variable Definitions:						*/

    long	bounds[2][8];	/* Output coordinates of input edges */ 
    long	i;		/* Loop counter/index */
    long	j;		/* Loop counter/index */
    long	k;		/* Loop counter/index */
    long	nlr;		/* Number of lines in current region */
    long	nsr;		/* Number of samples in region */
    float	sum;		/* Sum of weigths for one corner */
    float	wghts[4][4];	/* Combined weights at each corner of
				   each input image */
    long	xdist[4];	/* Number of pixels from upper left
				   corner of region to edge of input
				   image, in order NW, NE, SW, and SE */
    long	xdelta[4];	/* Corresponding change in distance with
				   increasing pixel coordinate */
    float	xwghts[4][4];	/* X-direction weights at each corner of
				   each input image */
    float	xsum;		/* Sum of x-direction weights */
    long	ydist[4];	/* Number of lines from upper left
				   corner of region to edge of input
				   image, in order NW, NE, SW, and SE;
				   -1 if image not within region */
    long	ydelta[4];	/* Corresponding change in distance with
				   increasing line coordinate */
    float	ywghts[4][4];	/* Y-direction weights at each corner of
				   each input image */
    float	ysum;		/* Sum of y-direction weights */

/*  Sort region terminating lines and samples N to S and W to E --
    When a new input window is encountered, the previous region is
    terminated at the last line or sample before the new window  */

    for (i = 0; i < 2; i++)
	{
	for (j = 0; j < 4; j++)
	    {
	    bounds[i][2*j] = inwind[j][i] - 1;
	    bounds[i][2*j+1] = inwind[j][i] + inwind[j][i+2] - 1;
	    }
	qsort ((void *) bounds[i], 8, sizeof (long), compare_edge);
	}

/*  Set boundaries of each region -- each region begins at the next line
    or sample after the end of the previous region  */

    for (i = 0; i < 49; i++)
	{
	region[i].inflags = 0;
	region[i].sls[0] = bounds[0][i/7] + 1;
	region[i].lls[0] = bounds[0][1+i/7];
	region[i].sls[1] = bounds[1][i%7] + 1;
	region[i].lls[1] = bounds[1][1+i%7];
	if ( (region[i].lls[0] < region[i].sls[0])
	  || (region[i].lls[1] < region[i].sls[1]) )
	    continue;

/*  Determine which input images contain region  */

	nlr = 1 + region[i].lls[0] - region[i].sls[0];
	nsr = 1 + region[i].lls[1] - region[i].sls[1];
	for (j = 0; j < 4; j++)
	    {
	    region[i].ulcwghts[j] = region[i].ystep[j]
	      = region[i].xstep[j] = region[i].dxstep[j] = 0.0;
	    if ( (inwind[j][0] <= region[i].sls[0])
	      && (inwind[j][1] <= region[i].sls[1])
	      && (inwind[j][0] + inwind[j][2] > region[i].lls[0])
	      && (inwind[j][1] + inwind[j][3] > region[i].lls[1]) )
		{
		region[i].inflags += 1 << j;

/*  Determine distances from region ulc to window edges  */

		ydist[j] = (j / 2) ?
		  region[i].sls[0] - inwind[j][0] + 1 :
		  inwind[j][0] + inwind[j][2] - region[i].sls[0];
		xdist[j] = (j % 2) ?
		  region[i].sls[1] - inwind[j][1]  + 1 :
		  inwind[j][1] + inwind[j][3] - region[i].sls[1];
		ydelta[j] = (j / 2) ? 1 : -1;
		xdelta[j] = (j % 2) ? 1 : -1;

/*  Compute relative weights for x and y directions one line and sample
    outside each corner  */


		ywghts[0][j] = ywghts[1][j] = ydist[j] - ydelta[j];
		ywghts[2][j] = ywghts[3][j]
		  = ydist[j] + ydelta[j] * nlr; 
		xwghts[0][j] = xwghts[2][j] = xdist[j] - xdelta[j];
		xwghts[1][j] = xwghts[3][j]
		  = xdist[j] + xdelta[j] * nsr;
		}
	    else
		for (k = 0; k < 4; k++)
		    ywghts[k][j] = xwghts[k][j] = 0;
	    }
	if (! region[i].inflags)
	    continue;

/*  Determine image weights just outside each corner or region  */

	for (k = 0; k < 4; k++)
	    {
	    ysum = xsum = sum = 0;
	    for (j = 0; j < 4; j++)
		{
		ysum += ywghts[k][j];
		xsum += xwghts[k][j];
		}

/*  Ignore different x or y weights if within image interior  */

	    if (ywghts[k][0] * ywghts[k][1] * xwghts[k][0] *
	      xwghts[k][1])
		{
		ywghts[k][0] = (ywghts[k][0] + ywghts[k][1]) / 2;
		ywghts[k][1] = ywghts[k][0];
		}
	    if (ywghts[k][2] * ywghts[k][3] * xwghts[k][2] *
	      xwghts[k][3])
		{
		ywghts[k][2] = (ywghts[k][2] + ywghts[k][3]) / 2;
		ywghts[k][3] = ywghts[k][2];
		}
	    if (ywghts[k][0] * ywghts[k][2] * xwghts[k][0] *
	      xwghts[k][2])
		{
		xwghts[k][0] = (xwghts[k][0] + xwghts[k][2]) / 2;
		xwghts[k][2] = xwghts[k][0];
		}
	    if (ywghts[k][1] * ywghts[k][3] * xwghts[k][1] *
	      xwghts[k][3])
		{
		xwghts[k][1] = (xwghts[k][1] + xwghts[k][3]) / 2;
		xwghts[k][3] = xwghts[k][1];
		}

/*  Compute normalized weight values  */

	    for (j = 0; j < 4; j++)
		{
		if (region[i].inflags & (1 << j))
		    {
		    ywghts[k][j] = (ysum > 0) ? ywghts[k][j] / ysum : 1;
		    xwghts[k][j] = (xsum > 0) ? xwghts[k][j] / xsum : 1;
		    wghts[k][j] = ywghts[k][j] * xwghts[k][j];
		    sum += wghts[k][j];
		    }
		else
		    wghts[k][j] = 0;
		}
	    if (sum <= 0)
		for (j = 0; j < 4; j++)
		    if (region[i].inflags & (1 << j))
			{
			wghts[k][j] = 1;
			sum++;
			}
	    for (j = 0; j < 4; j++)
		wghts[k][j] /= sum;
	    }

/*  Compute coefficients of bilinear weight interpolation -- the upper
    left corner of region is inset (1,1) from point for wghts[0][j]  */

	for (j = 0; j < 4; j++)
	    {
	    region[i].ystep[j] = (wghts[2][j] - wghts[0][j])
	      / (nlr + 1);
	    region[i].xstep[j] = (wghts[1][j] - wghts[0][j])
	      / (nsr + 1);
	    region[i].dxstep[j] = (wghts[0][j] - wghts[1][j] -
	      wghts[2][j] + wghts[3][j]) / ((nlr + 1) * (nsr + 1));
	    region[i].ulcwghts[j] = wghts[0][j] + region[i].ystep[j]
	      + region[i].xstep[j] + region[i].dxstep[j];
	    region[i].ystep[j] += region[i].dxstep[j];
	    region[i].xstep[j] += region[i].dxstep[j];
	    }

	}

    return E_SUCC;
    }			/* End of Function mrf_defrgn */
									/*
***************   Function compare_edge   ***************	     	

Purpose:
    Performs comparison needed for using qsort() to sort input image
    edges by output-image coordinate value.

*/

FUNCTION int compare_edge (
    const void	*edge1,		/* First edge for comparison */
    const void	*edge2)		/* Second edge for comparison */
    {
    long key1 = * (long *) edge1;
    long key2 = * (long *) edge2;

    return (key1 < key2) ? -1 : (key1 > key2) ? 1 : 0;

    }				/* End of function compare_edge */
