/*  mrf_getarg.c -- Get arguments for mergefour

NAME:
    mrf_getarg

PURPOSE:
    Called by MERGEFOUR to get arguments from TAE interface

ARGUMENTS:

    struct PARBLK
	    *vblock	(IN)	TAE parameter block.
    struct MRF_IMG
	    *input	(OUT)	Input image specifications.
    long    *inimgcnt	(OUT)	Number of input images.
    char    *outimg	(OUT)	Name of output image file.
    long    *odtype	(OUT)	Output data type code.
    long    *pmsgtim	(OUT)	Interval between progress messages.
    long    *ifminmax	(OUT)	Flag for output minima/maxima.

RETURN VALUE:  
    E_SUCC	Successful completion.
		(Function aborts on error)

PROGRAM HISTORY:                                  
    2000-05	Initial version, R. A. White, PSU/ESSC.
    2005-03	Unpdated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.X and Linux 2.4.

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in MERGEFOUR.PDF.

ALGORITHM DESCRIPTION:

    Allocate temporary arrays to receive input image descriptors
    Get names and sizes of input images
    Get output image name
    Get output data type
    Get other arguments
    Free temporory arrays
    									
***************   End of Prolog for mrf_getarg   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O definitions		*/
#include	<string.h>	/* String manipulation functions	*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"util.h"	/* LAS utility function definitions	*/
#include	"mergefour.h"	/* Program-specific definitions		*/
									/*
External Function Prototypes Not Included in Header Files:		*/
    struct VARIABLE *p_find (struct PARBLK *vptr, char *name);
									/*
***************   Start of Function mrf_getarg   ***************	*/

FUNCTION long mrf_getarg(

    struct PARBLK
	    *vblock,	/* TAE parameter block */
    struct MRF_IMG
	    *input,	/* Input image specifications */
    char    *outimg,	/* Name of output image file */
    long    *odtype,	/* Output data type code */
    long    *pmsgtim,	/* Interval between progress messages */
    long    *ifminmax)	/* Flag for output minima/maxima */
    {
	                                              			/*
Local Variable Definitions:						*/

    long	i;		/* Loop counter/index */
    long	imglimit;	/* Maximum images in logical image */
    long	j;		/* Loop counter/index */
    long	nli;		/* Number of lines in input image */
    long	nimg;		/* Number of images in logical image */
    long	nsi;		/* Number of samples per input line */
    char	odoption[]	/* Possible values of ODTYPE argument */
		  = "SAMEBYTEI*2 I*4 R*4";
    long	one = 1;	/* Constant one */
    long	qflag[MAXIMG][2];/* Flags for window defaults (ignored) */
    long	same_flag;	/* Flag for input windows all same */
    long	*tmpbnd;	/* Pointer into tmpbuf for bands */
    void	*tmpbuf;	/* Temporary buffer for image args */
    char	*tmphnam;	/* Pointer in tmpbuf for host names */
    long	*tmpnbnd;	/* Pointer into tmpbuf for nbands */
    long	*tmpwin;	/* Pointer into tmpbuf for window */
    char	*tmptnam;	/* Pointer into tmpbuf for TAE names */
    long	totband = 0;	/* Total bands in logical image */

/*  Allocate temporary buffer space for image specifications  */

    if ((tmpbuf = (void *) malloc (2 * MAXIMG * CMLEN
      + 4 * MAXIMG * sizeof (long)
      + MAXIMG * (MAXBND + 1) * sizeof (long)
      + MAXIMG * sizeof (long))) == NULL)
	c_lasmsg (TRUE, "mergefour-malloc", "Memory allocation error");
    tmpwin = (long *) tmpbuf;
    tmpbnd = tmpwin + 4 * MAXIMG;
    tmpnbnd = tmpbnd + MAXIMG * (MAXBND + 1);
    tmphnam = (char *) (tmpnbnd + MAXIMG);
    tmptnam = tmphnam + MAXIMG * CMLEN;

/*  Get names and sizes of input images  */

    imglimit = MAXIMG;
    nli = 0;
    for (i = 0; i < 4; i++)
	{
	if (c_getin (vblock, "IN", &i, (void *) tmptnam,
	  (void *) tmphnam, &(input[i].totimage), (void *) tmpwin,
	  (void *) tmpbnd, tmpnbnd, &(input[i].totband), &imglimit,
	  &same_flag, qflag)
	   != E_SUCC)
	    c_lasmsg (TRUE, "mergefour-fatal",
	      "Fatal error encountered");
	if (nli == 0)
	    {
	    nli = tmpwin[2];
	    nsi = tmpwin[3];
	    totband = (input)->totband;
	    }
	if (same_flag == 0)
	    c_lasmsg (TRUE, "mergefour-diffwin", "Images in logical"
	      " image have different window sizes");
	if (input[i].totband != totband)
	    c_lasmsg (TRUE, "mergefour-difband",
	      "Input images have different number of bands");

/*  Allocate array space in MRF_IMG structure for image  */

	nimg = input[i].totimage;
	if ( ((input[i].hname
	   = (char **) malloc (nimg * sizeof (char *))) == NULL)
	  || ((input[i].window
	   = (long *) malloc (nimg * 4 * sizeof (long *))) == NULL)
	  || ((input[i].nband
	   = (long *) malloc (nimg * sizeof (long))) == NULL)
	  || ((input[i].bands
	   = (long **) malloc (nimg * sizeof (long *))) == NULL)
	  || ((input[i].fdesc
	   = (void *) malloc ((nimg + 1) * sizeof (void *))) == NULL)
	  || ((input[i].offsets
	   = (long *) malloc (totband * sizeof (long))) == NULL) ) 
	    c_lasmsg (TRUE, "mergefour-malloc",
	      "Memory allocation error");

/*  Move arrays from temporary buffer to MRF_IMG structure for image  */

	for (j = 0; j < nimg; j++)
	    {
	    if ((input[i].hname[j]
	      = (char *) malloc (strlen (tmphnam + j * CMLEN) + 1))
	       == NULL)
		c_lasmsg (TRUE, "mergefour-malloc",
		  "Memory allocation error");
	    strcpy (input[i].hname[j], tmphnam + j * CMLEN);
	    memcpy (input[i].window + 4 * j, tmpwin + 4 * j,
	      4 * sizeof (long));
	    input[i].nband[j] = tmpnbnd[j];
	    if ((input[i].bands[j] = (long *)
	      malloc ((input[i].nband[j] + 1) * sizeof (long))) == NULL) 
		c_lasmsg (TRUE, "mergefour-malloc",
		  "Memory allocation error");
	    memcpy (input[i].bands[j], tmpbnd + j * (MAXBND + 1),
	      tmpnbnd[j] * sizeof (long));
	    }
	}

/*  Get name of output image  */

    if (c_getout (vblock, "OUT", (void *) tmptnam, (void *) outimg)
      != E_SUCC)
	c_lasmsg (TRUE, "mergefour-fatal", "Fatal error encountered");

/*  Get output data type and convert to data type code  */

    i = 5;
    if (c_getstr (vblock, "ODTYPE", &one, &i, (char *) tmpbuf) != one)
	c_lasmsg (TRUE, "mergefour-fatal", "Fatal error encountered");
    *odtype = (strstr (odoption, (char *) tmpbuf) - odoption) / 4;

/*  Get other arguments  */

    *pmsgtim = IVAL (*p_find (vblock, "$MSGTIM"), 0);
    strcpy ((char *) tmpbuf, SVAL (*p_find(vblock, "$MINMAX"), 0));
    if ( ((char *)tmpbuf)[0] == 'Y')
        *ifminmax = TRUE;
    else
        *ifminmax = FALSE;

/*  Free temporary buffer and return to caller  */

    free (tmpbuf);
    return E_SUCC;
    }			/* End of Function mrf_getarg */
