/*  mrf_getwnd.c -- Determine locations of inputs relative to output

NAME:
    mrf_getwnd

PURPOSE:
    Called by MERGEFOUR to establish the location of each input image
    relative to the output image.

ARGUMENTS:
    struct MRF_IMG
	    *input	(MOD)	Input image specifications.
    long    *inseq	(OUT)	Sequence for reading inputs.
    struct DDR
	    *outddr	(OUT)	DDR for output image.
    long    inwind[][4]	(OUT)	Input image window locations relative to
				output image.

RETURN VALUE:  
    E_SUCC	Successful completion.
		(Function aborts on error)

PROGRAM HISTORY:                                  
    2000-05	Initial version, R. A. White, PSU/ESSC.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.X and Linux 2.4.

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in MERGEFOUR.PDF.

ALGORITHM DESCRIPTION:

    Read DDRs for each input image
    Determine projection coordinates for each input window
    Identify NW, NE, SW, and SE inputs
    Find bounding projection coordinates of region covered by at least
      two input image windows
    Trim input image windows to fit within this region
    Find location of each input window relative to output
    Resequence input window specs in order NW, NE, SW, SE
    Initialize DDR for output image
    									
***************   End of Prolog for mrf_getwnd   ***************	*/
									/*
Header Files:			       					*/

#include	<math.h>	/* Mathematical functions		*/
#include	<stdio.h>	/* Standard I/O definitions		*/
#include	<string.h>	/* String manipulation functions	*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"util.h"	/* LAS utility function definitions	*/
#include	"mergefour.h"	/* Program-specific definitions		*/
									/*
***************   Start of Function mrf_getwnd   ***************	*/

FUNCTION long mrf_getwnd(

    struct MRF_IMG
	    *input,	/* Input image specifications */
    long    *inseq,	/* Sequence for reading inputs */
    struct DDR
	    *outddr,	/* DDR for output image */
    long    inwind[][4]) /* Input image window locations relative to
			   output image */
    {
	                                              			/*
Local Variable Definitions:						*/
    long	above;		/* Number of inputs a given input's top
				   is above */
    double	coords[4][4];	/* Coords for each input window in order
				   uly, ulx, lry, lrx */
    long	delta;		/* Change in input image coodinate */
    struct DDR	*inddr;		/* Input image DDRs */
    long	geowarn;	/* Flag set if geoloc warning issued */
    long	i;		/* Loop counter/index */
    long	j;		/* Loop counter/index */
    long	k;		/* Loop counter/index */
    long	left;		/* Number of inputs a given input's left
				   edge is left of */
    long	len;		/* Length of character string */
    double	maxcoord[2];	/* Maximum input window coords */ 
    double	mincoord[2];	/* Minimum input window coords */ 
    long	nlout;		/* Number of output lines */
    long	nsout;		/* Number of samples per output line */

/*  Read DDRs for each input image  */

    if ((inddr = (struct DDR *) malloc (5 * sizeof (struct DDR)))
      == NULL) 
	c_lasmsg (TRUE, "mergefour-malloc", "Memory allocation error");
    for (i = 0, geowarn = 0; i < 4; i++)
	{
	if (c_getddr (input[i].hname[0], inddr + i) != E_SUCC)
	    c_lasmsg (TRUE, "mergefour-fatal",
	      "Fatal error encountered");
	len = strlen (inddr[i].proj_units);
	c_up2low (inddr[i].proj_units, &len);
	if (! strcmp (inddr[i].proj_units, "dms"))
	    {
	    convert_ddr (inddr + i, "DEG");
	    strcpy (inddr[i].proj_units, "deg");
	    }
	for (j =1; (j < input[i].totimage) && (! geowarn); j++)
	    {
	    if (c_getddr (input[i].hname[j], inddr + 4) != E_SUCC)
		c_lasmsg (TRUE, "mergefour-fatal",
		  "Fatal error encountered");
	    len = strlen (inddr[4].proj_units);
	    c_up2low (inddr[4].proj_units, &len);
	    if (! strcmp (inddr[4].proj_units, "dms"))
		{
		convert_ddr (inddr + 4, "DEG");
		strcpy (inddr[4].proj_units, "DEG");
		}
	    if ( (inddr[i].valid[DDPCV] == VALID)
	      && (inddr[4].valid[DDPCV] == VALID)
	      && (inddr[i].valid[DDPDV] == VALID)
	      && (inddr[4].valid[DDPDV] == VALID)
	      && (inddr[i].valid[DDPUV] == VALID)
	      && (inddr[4].valid[DDPUV] == VALID)
	      && (inddr[i].valid[DDCCV] == VALID)
	      && (inddr[4].valid[DDCCV] == VALID) )
		{
		if ( (inddr[i].proj_code != inddr[4].proj_code)
		  || (strcmp (inddr[i].proj_units,
		       inddr[4].proj_units)) 
		  || (inddr[i].pdist_y != inddr[4].pdist_y)
		  || (inddr[i].pdist_x != inddr[4].pdist_x)
		  || (fabs ( ((inddr[i].upleft[0] - inddr[4].upleft[0])
		    / inddr[i].pdist_y) + (input[i].window[0]
		    - input[i].window[4*j])) > 0.5)
		  || (fabs ( ((inddr[i].upleft[1] - inddr[4].upleft[1])
		    / inddr[i].pdist_x) - (input[i].window[1]
		    - input[i].window[4*j+1])) > 0.5) )
		    {
		    c_lasmsg (FALSE, "mergefour-geoloc", "Geographic"
		      " coordinates for input %s window don't match"
		      " those for %s", input[i].hname[0],
		      input[i].hname[j]);
		    geowarn = 1;
		    break;
		    }
		}
	    else
		{
		c_lasmsg (FALSE, "mergefour-geoloc", "Invalid projection"
		  " data preclude verifying geographic registstration"
		  " for %s and %s", input[i].hname[0],
		  input[i].hname[j]);
		geowarn = 1;
		break;
		}
	    }
	}

/*  Determine projection coordinates for each input window  */

    for (i = 0, geowarn = 0; i < 4; i++)
	{
	if (! inddr[i].valid[DDPCV])
	    {
	    if (! geowarn++)
		c_lasmsg (FALSE, "mergefour-projver", "Invalid"
	      " projection data in DDR preclude verifying that"
	      " projections are same for all inputs");
	    }
	else if ((i > 0) && (inddr[i].proj_code != inddr[0].proj_code)
	  && (! geowarn))
	    c_lasmsg (TRUE, "mergefour-difproj", "Inputs %s and %s"
	      " are in different projections", input[i].hname[0],
	      input[0].hname[0]);
	else if ((i > 0) && (strcmp (inddr[i].proj_units,
	  inddr[0].proj_units)) && (! geowarn))
	    c_lasmsg (TRUE, "mergefour-difunit", "Inputs %s and %s"
	      " use different projection units", input[i].hname[0],
	      input[0].hname[0]);
	if ( (! inddr[i].valid[DDPDV]) || (! inddr[i].valid[DDCCV]))
	    c_lasmsg (TRUE, "mergefour-invcorn", "Input %s does not"
	      " have valid entry for grid spacing or corner coords",
	      input[i].hname[0]);
	if ( (i > 0) && ((inddr[i].pdist_x != inddr[0].pdist_x)
	  || (inddr[i].pdist_y != inddr[0].pdist_y)) )
	    c_lasmsg (TRUE, "mergefour-difgrid", "Inputs %s and %s"
	      " have different grid spacings", input[i].hname[0],
	      input[0].hname[0]);
	coords[i][0] = inddr[i].upleft[0]
	  - inddr[i].pdist_y * (input[i].window[0] - 1);
	coords[i][1] = inddr[i].upleft[1]
	  + inddr[i].pdist_x * (input[i].window[1] - 1);
	coords[i][2] = coords[i][0]
	  - inddr[i].pdist_y * (input[i].window[2] - 1);
	coords[i][3] = coords[i][1]
	  + inddr[i].pdist_x * (input[i].window[3] - 1);
	}

/*  Identify NW, NE, SW, and SE inputs  */

    for (i = 0; i < 4; i++)
	inseq[i] = -1;
    for (i = 0; i < 4; i++)
	{
	above = left = 0;
	for (j = 0; j < 4; j++)
	    {
	    if (j == i)
		continue;
	    if (coords[i][0] != coords[j][0])
		above += (coords[i][0] > coords[j][0]) ? 3 : -3;
	    else
		above +=  (i < j) ? 1 : -1; 
	    if (coords[i][2] != coords[j][2])
		above += (coords[i][2] > coords[j][2]) ? 3 : -3;
	    else
		above +=  (i < j) ? 1 : -1; 
	    if (coords[i][1] != coords[j][1])
		left += (coords[i][1] < coords[j][1]) ? 3 : -3;
	    else
		left +=  (i < j) ? 1 : -1; 
	    if (coords[i][3] != coords[j][3])
		left += (coords[i][3] < coords[j][3]) ? 3 : -3;
	    else
		left +=  (i < j) ? 1 : -1; 
	    }
	k = (above >= 0) ? ((left >= 0) ? 0 : 1)
	  : ((left >= 0) ? 2 : 3);

/*  Allow for identical inputs when overlap in only one direction --
    this may fail depending on order of inputs  */

	if ((inseq[k] >= 0)
	  && (! memcmp (coords[i], coords[inseq[k]],
	   4 * sizeof (double))))  
	    k = (abs(above) >= abs(left)) ? (5 - k) % 4 : (k + 2) % 4;
	inseq[k] = i;
	}
    for (k = 0; k < 4; k++)
	if (inseq[k] < 0)
	    break;
    if (k < 4)
	c_lasmsg (TRUE, "mergefour-locate", "Cannot assign inputs to"
	  " NW, NE, SW, SE locations");

/*  Find bounding projection coordinates of region covered by at least
    two input image windows  */

    mincoord[0] = (coords[inseq[2]][2] > coords[inseq[3]][2]) ?
      coords[inseq[2]][2] : coords[inseq[3]][2];
    mincoord[1] = (coords[inseq[0]][1] > coords[inseq[2]][1]) ?
      coords[inseq[0]][1] : coords[inseq[2]][1];
    maxcoord[0] = (coords[inseq[0]][0] < coords[inseq[1]][0]) ?
      coords[inseq[0]][0] : coords[inseq[1]][0];
    maxcoord[1] = (coords[inseq[1]][3] < coords[inseq[3]][3]) ?
      coords[inseq[1]][3] : coords[inseq[3]][3];
    nlout
      = floor ((maxcoord[0] - mincoord[0]) / inddr[0].pdist_y + 1.5); 
    mincoord[0] = maxcoord[0] - inddr[0].pdist_y * (nlout - 1);
    nsout
      =  floor ((maxcoord[1] - mincoord[1]) / inddr[0].pdist_x + 1.5);
    maxcoord[1] = mincoord[1] + inddr[0].pdist_x * (nsout - 1);

/*  Trim input image windows to fit within this region  */

    for (i = 0; i < 4; i++)
	{
	if ((delta = ceil (- 0.5 + (coords[i][0] - maxcoord[0]) /
	  inddr[0].pdist_y)) > 0)
	    {
	    input[i].window[0] += delta;
	    coords[i][0] -= delta * inddr[0].pdist_y;
	    }
	if ((delta = floor (0.5 + (mincoord[0] - coords[i][2]) /
	  inddr[0].pdist_y)) > 0)
	    coords[i][2] += delta * inddr[0].pdist_y;
	if ((delta = ceil (-0.5 + (mincoord[1] - coords[i][1]) /
	  inddr[0].pdist_x)) > 0)
	    {
	    input[i].window[1] += delta;
	    coords[i][1] += delta * inddr[0].pdist_x;
	    }
	if ((delta = floor (0.5 + (coords[i][3] - maxcoord[1]) /
	  inddr[0].pdist_x)) > 0)
	    coords[i][3] -= delta * inddr[0].pdist_x;
	}

/*  Find output location of each resequenced input window  */

    for (i = 0; i < 4; i++)
	{
	inwind[i][0] = floor (1.5 + (maxcoord[0] - coords[inseq[i]][0])
          / inddr[0].pdist_y);
	inwind[i][1] = floor (1.5 + (coords[inseq[i]][1] - mincoord[1])
	  / inddr[0].pdist_x);
	inwind[i][2] = floor (2.5 + (maxcoord[0] - coords[inseq[i]][2])
          / inddr[0].pdist_y) - inwind[i][0];
	input[inseq[i]].window[2] = inwind[i][2];
	inwind[i][3] = floor (2.5 + (coords[inseq[i]][3] - mincoord[1])
	  / inddr[0].pdist_x) - inwind[i][1];
	input[inseq[i]].window[3] = inwind[i][3];
	if ( (inwind[i][0] < 1) || (inwind[i][1] < 1)
	  || (inwind[i][2] > inddr[inseq[i]].nl)
	  || (inwind[i][3] > inddr[inseq[i]].ns) )
	    c_lasmsg (TRUE, "mergefour-inwind", "Input window"
	      " specification for %s extends outside image",
	      input[inseq[i]].hname[0]);
	for (j = 1; j < input[i].totimage; j++)
	    for (k = 0; k < 4; k++)
		input[i].window[4*j+k] = input[i].window[k];
	}

/*  Initialize DDR for output image  */

    *outddr = inddr[0];
    if (geowarn)
	outddr->valid[DDPCV] = INVAL;
    outddr->nl = nlout;
    outddr->ns = nsout;
    outddr->dtype = inddr[0].dtype;
    strcpy (outddr->proj_units, inddr[0].proj_units);
    outddr->upleft[0] = outddr->upright[0] = maxcoord[0];
    outddr->upleft[1] = outddr->loleft[1] = mincoord[1];
    outddr->loleft[0] = outddr->loright[0] = outddr->upleft[0]
      - (outddr->nl - 1) * outddr->pdist_y;
    outddr->upright[1] = outddr->loright[1] = outddr->upleft[1]
      + (outddr->ns - 1) * outddr->pdist_x;

    free (inddr);
    return E_SUCC;
    }			/* End of Function mrf_getwnd */
