/*  mrf_merge.c -- Merge input images

NAME:
    mrf_merge

PURPOSE:
    Called by MERGEFOUR to merge the pixels from the four input images.

ARGUMENTS:
    struct MRF_IMG
	    *input	(IN)	Input image specifications.
    long    pmsgtim	(IN)	Interval between progress messages.
    long    *inseq	(IN)	Sequence for reading inputs.
    long    inwind[][4]	(IN)	Input image window locations relative to
				output image.
    struct MRF_REGION
	    *region	(IN)	Specifications for each overlap region.
    char    *outimg	(IN)	Name of output image.
    struct DDR
	    *outddr	(IN)	DDR for output image.
    long    *odtype	(MOD)	Output data type code.

RETURN VALUE:  
    E_SUCC	Successful completion.
    E_FAIL	Error occurred after actual processing began -- function
		aborts if error occurs during initialization.

PROGRAM HISTORY:                                  
    2000-05	Initial version, R. A. White, PSU/ESSC.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:
    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.X and Linux 2.4.

PROJECT:
    LAS at PSU/ESSC

INFORMATIONAL AND ERROR MESSAGES:

    These are described in MERGEFOUR.PDF.

ALGORITHM DESCRIPTION:
    Open input images as separate image groups and allocate buffers
    Open output image as image group and allocate buffer
    Initialize processing progress messages
    Initialize min/max determination
    For each row of overlap regions
        Determine which inputs are used for this row
	For each output line
	    Initialize output accumulators
	    For each region in row
		For each pixel
		    For each band
			Compute weighted average of input pixels
		    Increment weights for next pixel
	    Update min/max values
	    Write output line
	    Read next line for each contributing input image
	    Increment weights and steps to new line
    Close inputs and outputs
    Stop processing progress messages
    Merge georeferencing information into output DDR
    Save minima and maxima
	
    									
***************   End of Prolog for mrf_merge   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O definitions		*/
#include 	<string.h>	/* Standard string processing functions	*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"pixman.h"	/* LAS pixel maniputaltion functions	*/
#include	"util.h"	/* LAS utility function definitions	*/
#include	"mergefour.h"	/* Program-specific definitions		*/
									/*
External Variable Declarations:						*/
    extern int curent;		/* Entity for progress messages */
									/*
***************   Start of Function mrf_merge   ***************	*/

FUNCTION long mrf_merge(

    struct MRF_IMG
	    *input,	/* Input image specifications */
    long    pmsgtim,	/* Interval between progress messages */
    long    *inseq,	/* Sequence for reading inputs */
    long    inwind[][4], /* Input image window locations relative to
			   output image */
    struct MRF_REGION
	    *region,	/* Specifications for each overlap region */
    char    *outimg,	/* Name of output image */
    struct DDR
	    *outddr,	/* DDR for output image */
    long    *odtype)	/* Output data type code */

    {
	                                              			/*
Local Variable Definitions:						*/

    long	access;		/* Access code for file open */
    long	band;		/* Current band number */
    struct BDDR *bddr = NULL;	/* Temporary storage for BDR data */
    long	buffersz;	/* Size of image buffer */
    long	curregion;	/* Current region */
    long	idcode;		/* Data-type code for input buffers */
    long	i;		/* Loop counter/index */
    double	*inbuf[4];	/* Input buffers */
    struct GDESC
		*ingdesc[4];	/* Input group descriptors */
    long	inimg;		/* Current input image */
    long	*inoffs[4];	/* Locations of input offsets arrays */
    struct MRF_IMG
		*inspec;	/* Specifiers for current input */
    long	line;		/* Current output line */
    double	maxlegal;	/* Maximum legal value for output  */ 
    double	*maxval;	/* Maximum values for output bands */ 
    double	minlegal;	/* Minimum legal value for output  */ 
    double	*minval;	/* Minimum values for output bands */ 
    long	nband;		/* Number of bands in output image */
    struct DDR	newddr;		/* Newly created output ddr */
    long	one = 1;	/* Constant value */
    long	opt[2] = {0, 0};/* Option codes for file open */
    double	*outbuf;	/* Output buffer */
    struct FDESC
		*outfd[2];	/* Output file descriptor */
    struct GDESC
		*outgdesc;	/* Output group descriptor */
    long	*outoffs;	/* Output-buffer offsets array */
    double	outval;		/* Value of current output pixel */
    long	row;		/* Current row of regions */
    long	samp;		/* Current sample position in output */
    long	used[4];	/* Flags set for used for row */
    float	wghts[4];	/* Current weights for each input */

/*  Open input images as separate image groups  */

    access = IREAD;
    idcode = EDOUBLE;
    for (inimg = 0; inimg < 4; inimg++)
	{
	inspec = input + inseq[inimg];
	inoffs[inimg] = inspec->offsets;
	for (i = 0, band = 0; i < inspec->totimage;
	   band += inspec->nband[i], i++)
	    {
	    inspec->bands[i][inspec->nband[i]] = 0;
	    if (c_eopens (&inspec->fdesc[i], inspec->hname[i],
	      inspec->bands[i], &inspec->window[4*i],
	      &inspec->window[4*i+1], &inspec->window[4*i+2],
	      &inspec->window[4*i+3], &inoffs[inimg][band], &access,
	      &idcode, opt, opt) != E_SUCC)
		c_lasmsg (TRUE, "mergefour-fatal",
		  "Fatal error encountered");
	    }
	inspec->fdesc[i] = 0;
	if ((inbuf[inimg] = (double *) c_ealloc (inspec->fdesc,
	  &buffersz)) == NULL)
	    c_lasmsg (TRUE, "mergefour-fatal",
	      "Fatal error encountered"); 
	if (c_egroup (inspec->fdesc, &ingdesc[inimg],
	  (void *) inbuf[inimg], &buffersz) != E_SUCC)
	    c_lasmsg (TRUE, "mergefour-fatal",
	      "Fatal error encountered");
	if (c_estep (&ingdesc[inimg]) != E_SUCC)
	    c_lasmsg (TRUE, "mergefour-fatal",
	      "Fatal error encountered");
	}

/*  Open output image as image group and allocate buffer  */

    if (*odtype == 0)
	*odtype = outddr->dtype;
    nband = input[0].totband;
    if ((outoffs = (long *) malloc (nband * sizeof (long))) == NULL)
	c_lasmsg (TRUE, "mergefour-malloc", "Memory allocation error");
    access = IWRITE;
    opt[0] = *odtype;
    if (c_eopens (outfd, outimg, &nband, &one, &one, &outddr->nl,
      &outddr->ns, outoffs, &access, &idcode, &opt[1], opt) != E_SUCC) 
	c_lasmsg (TRUE, "mergefour-fatal", "Fatal error encountered");
    outfd[1] = 0;
    if ((outbuf = (double *) c_ealloc (outfd, &buffersz)) == NULL)
	c_lasmsg (TRUE, "mergefour-fatal", "Fatal error encountered");
    if (c_egroup (outfd, &outgdesc, (void *) outbuf, &buffersz)
      != E_SUCC)
	c_lasmsg (TRUE, "mergefour-fatal", "Fatal error encountered");
    if (c_estep (&outgdesc) != E_SUCC)
	c_lasmsg (TRUE, "mergefour-fatal", "Fatal error encountered");

/*  Initialize min/max determination  */

    if ( ((minval = (double *) malloc (2 * nband * sizeof (double)) )
        == NULL)
      || ((bddr = (struct BDDR *) malloc (nband * sizeof (struct BDDR)))
       == NULL) )
	c_lasmsg (TRUE, "mergefour-malloc", "Memory allocation error");
    maxval = minval + nband;
    if (c_inmnmx (odtype, &nband, minval, maxval) != E_SUCC)
	c_lasmsg (TRUE, "mergefour-fatal", "Fatal error encountered");
    maxlegal = minval[0];
    minlegal = maxval[0];

/*  Initialize processing progress messages  */

    if (pmsgtim > 0)
	{
	curent = 0;
	i = GEN2;
	c_inittm ("MERGEFOUR", "processing", "line", "", "", &outddr->nl,
	  &nband, &pmsgtim, &i);
	c_startm ();
	}

/*  Process one row of overlap regions at a time  */

    for (row = 0; row < 7; row++)
	{
	if (region[7*row].sls[0] > region[7*row].lls[0])
	    continue;

/*  Determine which inputs are used for this row  */

	for (inimg = 0; inimg < 4; inimg++)
	    {
	    for (curregion = 7 * row; (curregion < 7 * (row + 1))
	       && (region[curregion].ulcwghts[inimg] <= 0.0);
	      curregion++);
	    used[inimg] = (curregion < 7 * (row + 1)) ? 1 : 0;
	    }

/*  Process each line in current row  */

	for (line = region[7*row].sls[0]; line <= region[7*row].lls[0];
	  line++)
	    {
	    curent = line;
	    for (samp = 0; samp < outddr->ns; samp++)
		for (band = 0; band < nband; band++)
		    outbuf[outoffs[band] + samp] = 0.0;

/*  Compute weighted average of inputs for each sample in each region  */

	    for (curregion = 7 * row; curregion < 7 * (row + 1);
	      curregion++)
		{
		if (region[curregion].sls[1] > region[curregion].lls[1])
		    continue;
		for (inimg = 0; inimg < 4; inimg++)
		    wghts[inimg] = region[curregion].ulcwghts[inimg];		
		for (samp = region[curregion].sls[1] - 1;
		  samp < region[curregion].lls[1]; samp++)
		    {
		    for (inimg = 0; inimg < 4; inimg++)
			{
			if (wghts[inimg] <= 0.0)
			    continue;
			for (band = 0; band < nband; band++)
			    outbuf[outoffs[band] + samp] += wghts[inimg]
			      * inbuf[inimg][inoffs[inimg][band]++];
			wghts[inimg] += region[curregion].xstep[inimg];
			}
		    for (band = 0; band < nband; band++)
			{
			outval = outbuf[outoffs[band] + samp];
			outval += (outval >= -0.5) ? 0.5 : -0.5;
			if (outval > maxlegal)
			    outval = maxlegal;
			if (outval < minlegal)
			    outval = minlegal;
			outbuf[outoffs[band] + samp] = outval;
			}
		    }
		}

/*  Update min/max values  */
	    
	    for (band = 0; band < nband; band++)
		if (c_pxmnmx ((void *) (outbuf + outoffs[band]),
		   &outddr->ns, minval + band, maxval + band, &idcode)
		  != E_SUCC)
		    {
		    c_egclse (&outgdesc);
		    return E_FAIL;
		    }

/*  Write output line and read next inputs  */

	    if (c_estep (&outgdesc) != E_SUCC)
		{
		c_egclse (&outgdesc);
		return E_FAIL;
		}
	    for (inimg = 0; inimg < 4; inimg++)
		if ((used[inimg]) && (line < outddr->nl)
		  && (c_estep (ingdesc + inimg)) != E_SUCC)
		    {
		    c_egclse (&outgdesc);
		    return E_FAIL;
		    }

/*  Increment weights and steps to new line  */

	    if (line == region[7*row].lls[0])
		break;
	    for (curregion = 7 * row; curregion < 7 * (row + 1);
	      curregion++)
		{
		if (region[curregion].sls[1] > region[curregion].lls[1])
		    continue;
		for (inimg = 0; inimg < 4; inimg++)
		    {
		    region[curregion].ulcwghts[inimg]
		      += region[curregion].ystep[inimg];		
		    region[curregion].xstep[inimg]
		      += region[curregion].dxstep[inimg];
		    }
		}
	    }

	}

/*  Stop processing messages and close files  */

    if (pmsgtim > 0)
	c_stoptm ();
    for (inimg = 0; inimg < 4; inimg++)
	c_egclse (ingdesc + inimg);
    c_egclse (&outgdesc);

/*  Merge georeferencing information into output DDR  */

    if (c_getddr (outimg, &newddr) != E_SUCC)
	return E_FAIL;
    for (band = 0, inimg = 0; inimg < input[0].totimage; inimg++)
	for (i = 0; i < input[0].nband[inimg]; i++)
	    if (c_getbdr (input[0].hname[inimg], &bddr[band++],
	      &input[0].bands[inimg][i]) != E_SUCC)
		{
		c_lasmsg (FALSE, "mergefour-noupbdr",
		  "Unable to update output BDDR");
		break;
		}
    memcpy (newddr.valid, outddr->valid, 8 * sizeof (long));
    newddr.proj_code = outddr->proj_code;
    newddr.zone_code = outddr->zone_code;
    newddr.datum_code = outddr->datum_code;
    strcpy (newddr.proj_units, outddr->proj_units);
    memcpy (newddr.proj_coef, outddr->proj_coef, 15 * sizeof (double));
    memcpy (newddr.upleft, outddr->upleft, 8 * sizeof (double));
    newddr.pdist_y = outddr->pdist_y;
    newddr.pdist_x = outddr->pdist_x;
    newddr.line_inc = outddr->line_inc;
    newddr.sample_inc = outddr->sample_inc;
    if (c_putddr (outimg, &newddr) != E_SUCC)
        {
        c_lasmsg (FALSE, "mergefour-noupddr",
	  "Unable to update output DDR");
        return E_SUCC;
        }
    for (i = 0; i < newddr.nbands; i++)
        {
	bddr[i].bandno = i + 1;
        if (c_putbdr (outimg, &bddr[i]) != E_SUCC)
            {
	    c_lasmsg (FALSE, "mergefour-noupbdr",
	      "Unable to update output BDDR");
	    return E_SUCC;
	    }
	}
      
/*  Save minima and maxima  */

    if (c_svmnmx (outimg, &newddr.nbands, minval, maxval, "YES")
      != E_SUCC)
	c_lasmsg (FALSE, "mergefour-noupbdr",
	  "Unable to update output BDDR");

    return E_SUCC;
    }			/* End of Function mrf_merge */
